"""
Module for serialization and deserialization of KNX/IP CEMI Frame.

cEMI stands for Common External Message Interface

A CEMI frame is the container to transport a KNX/IP Telegram to/from the KNX bus.

Documentation within:

    Application Note 117/08 v02
    KNX IP Communication Medium
    File: AN117 v02.01 KNX IP Communication Medium DV.pdf
"""
from xknx.exceptions import ConversionError, CouldNotParseKNXIP
from xknx.knx import (
    DPTArray, DPTBinary, GroupAddress, PhysicalAddress, Telegram, TelegramType)

from .body import KNXIPBody
from .knxip_enum import APCICommand, CEMIFlags, CEMIMessageCode


class CEMIFrame(KNXIPBody):
    """Representation of a CEMI Frame."""

    # pylint: disable=too-many-instance-attributes

    def __init__(self, xknx):
        """Initialize CEMIFrame object."""
        super(CEMIFrame, self).__init__(xknx)
        self.code = CEMIMessageCode.L_DATA_IND
        self.flags = 0
        self.cmd = APCICommand.GROUP_READ
        self.src_addr = GroupAddress(None)
        self.dst_addr = GroupAddress(None)
        self.mpdu_len = 0
        self.payload = None

    @property
    def telegram(self):
        """Return telegram."""
        telegram = Telegram()
        telegram.payload = self.payload
        telegram.group_address = self.dst_addr

        def resolve_telegram_type(cmd):
            """Return telegram type from APCI Command."""
            if cmd == APCICommand.GROUP_WRITE:
                return TelegramType.GROUP_WRITE
            if cmd == APCICommand.GROUP_READ:
                return TelegramType.GROUP_READ
            if cmd == APCICommand.GROUP_RESPONSE:
                return TelegramType.GROUP_RESPONSE
            raise ConversionError("Telegram not implemented for {0}".format(self.cmd))

        telegram.telegramtype = resolve_telegram_type(self.cmd)

        # TODO: Set telegram.direction [additional flag within KNXIP]
        return telegram

    @telegram.setter
    def telegram(self, telegram):
        """Set telegram."""
        self.dst_addr = telegram.group_address
        self.payload = telegram.payload

        # TODO: Move to separate function, together with setting of
        # CEMIMessageCode
        self.flags = (CEMIFlags.FRAME_TYPE_STANDARD |
                      CEMIFlags.DO_NOT_REPEAT |
                      CEMIFlags.BROADCAST |
                      CEMIFlags.PRIORITY_LOW |
                      CEMIFlags.NO_ACK_REQUESTED |
                      CEMIFlags.CONFIRM_NO_ERROR |
                      CEMIFlags.DESTINATION_GROUP_ADDRESS |
                      CEMIFlags.HOP_COUNT_1ST)

        # TODO: use telegram.direction
        def resolve_cmd(telegramtype):
            """Resolve APCICommand from TelegramType."""
            if telegramtype == TelegramType.GROUP_READ:
                return APCICommand.GROUP_READ
            if telegramtype == TelegramType.GROUP_WRITE:
                return APCICommand.GROUP_WRITE
            if telegramtype == TelegramType.GROUP_RESPONSE:
                return APCICommand.GROUP_RESPONSE
            raise TypeError()
        self.cmd = resolve_cmd(telegram.telegramtype)

    def set_hops(self, hops):
        """Set hops."""
        # Resetting hops
        self.flags &= 0xFFFF ^ 0x0070
        # Setting new hops
        self.flags |= hops << 4

    def calculated_length(self):
        """Get length of KNX/IP body."""
        if self.payload is None:
            return 11
        if isinstance(self.payload, DPTBinary):
            return 11
        if isinstance(self.payload, DPTArray):
            return 11 + len(self.payload.value)
        raise TypeError()

    def from_knx(self, raw):
        """Parse/deserialize from KNX/IP raw data."""
        self.code = CEMIMessageCode(raw[0])

        if self.code == CEMIMessageCode.L_DATA_IND or \
                self.code == CEMIMessageCode.L_Data_REQ or \
                self.code == CEMIMessageCode.L_DATA_CON:
            return self.from_knx_data_link_layer(raw)
        raise CouldNotParseKNXIP("Could not understand CEMIMessageCode: {0} / {1}".format(self.code, raw[0]))

    def from_knx_data_link_layer(self, cemi):
        """Parse L_DATA_IND, CEMIMessageCode.L_Data_REQ, CEMIMessageCode.L_DATA_CON."""
        if len(cemi) < 11:
            raise CouldNotParseKNXIP("CEMI too small")

        # AddIL (Additional Info Length), as specified within
        # KNX Chapter 3.6.3/4.1.4.3 "Additional information."
        # Additional information is not yet parsed.
        addil = cemi[1]

        self.flags = cemi[2 + addil] * 256 + cemi[3 + addil]

        self.src_addr = PhysicalAddress((cemi[4 + addil], cemi[5 + addil]))

        if self.flags & CEMIFlags.DESTINATION_GROUP_ADDRESS:
            self.dst_addr = GroupAddress((cemi[6 + addil], cemi[7 + addil]),
                                         levels=self.xknx.address_format)
        else:
            self.dst_addr = PhysicalAddress((cemi[6 + addil], cemi[7 + addil]))

        self.mpdu_len = cemi[8 + addil]

        # TPCI (transport layer control information)   -> First 14 bit
        # APCI (application layer control information) -> Last  10 bit

        tpci_apci = cemi[9 + addil] * 256 + cemi[10 + addil]

        self.cmd = APCICommand(tpci_apci & 0xFFC0)

        apdu = cemi[10 + addil:]
        if len(apdu) != self.mpdu_len:
            raise CouldNotParseKNXIP(
                "APDU LEN should be {} but is {}".format(
                    self.mpdu_len, len(apdu)))

        if len(apdu) == 1:
            apci = tpci_apci & DPTBinary.APCI_BITMASK
            self.payload = DPTBinary(apci)
        else:
            self.payload = DPTArray(cemi[11 + addil:])

        return 10 + addil + len(apdu)

    def to_knx(self):
        """Serialize to KNX/IP raw data."""
        if not isinstance(self.src_addr, (GroupAddress, PhysicalAddress)):
            raise ConversionError("src_add not set")
        if not isinstance(self.dst_addr, (GroupAddress, PhysicalAddress)):
            raise ConversionError("dst_add not set")

        data = []

        data.append(self.code.value)
        data.append(0x00)
        data.append((self.flags >> 8) & 255)
        data.append(self.flags & 255)
        data.extend(self.src_addr.to_knx())
        data.extend(self.dst_addr.to_knx())

        def encode_cmd_and_payload(cmd, encoded_payload=0, appended_payload=None):
            """Encode cmd and payload."""
            if appended_payload is None:
                appended_payload = []
            data = [
                1 + len(appended_payload),
                (cmd.value >> 8) & 0xff,
                (cmd.value & 0xff) |
                (encoded_payload & DPTBinary.APCI_BITMASK)]
            data.extend(appended_payload)
            return data

        if self.payload is None:
            data.extend(encode_cmd_and_payload(self.cmd))
        elif isinstance(self.payload, DPTBinary):
            data.extend(encode_cmd_and_payload(
                self.cmd,
                encoded_payload=self.payload.value))
        elif isinstance(self.payload, DPTArray):
            data.extend(encode_cmd_and_payload(
                self.cmd,
                appended_payload=self.payload.value))
        else:
            raise TypeError()
        return data

    def __str__(self):
        """Return object as readable string."""
        return '<CEMIFrame SourceAddress="{0}" DestinationAddress="{1}" ' \
               'Flags="{2:16b}" Command="{3}" payload="{4}" />'.format(
                   self.src_addr.__repr__(),
                   self.dst_addr.__repr__(),
                   self.flags,
                   self.cmd,
                   self.payload)

    def __eq__(self, other):
        """Equal operator."""
        return self.__dict__ == other.__dict__



"""Implementation of Basic KNX 2-Byte/octet values."""

from xknx.exceptions import ConversionError

from .dpt import DPTBase


class DPT2ByteUnsigned(DPTBase):
    """
    Abstraction for KNX 2 Byte "2-octet unsigned value".

    Contains smaller counters, timers  etc.

    DPT 7.xxx
    """

    value_min = 0
    value_max = 65535
    unit = ""
    resolution = 1
    payload_length = 2

    @classmethod
    def from_knx(cls, raw):
        """Parse/deserialize from KNX/IP raw data."""
        cls.test_bytesarray(raw, 2)
        return (raw[0] * 256) + raw[1]

    @classmethod
    def to_knx(cls, value):
        """Serialize to KNX/IP raw data."""
        if not isinstance(value, (int)):
            raise ConversionError("Cant serialize DPT2ByteUnsigned", value=value)
        if not cls._test_boundaries(value):
            raise ConversionError("Cant serialize DPT2ByteUnsigned", value=value)
        return value >> 8, value & 0xff

    @classmethod
    def _test_boundaries(cls, value):
        """Test if value is within defined range for this object."""
        return cls.value_min <= value <= cls.value_max


class DPT2Ucount(DPT2ByteUnsigned):
    """DPT 7.001 DPT_Value_2_Ucount."""

    unit = "pulses"


class DPTUElCurrentmA(DPT2ByteUnsigned):
    """DPT 7.012 Abstraction for KNX 2 Byte DPTUElCurrentmA."""

    unit = "mA"


class DPTBrightness(DPT2ByteUnsigned):
    """DPT 7.013 DPT_Brightness (lux)."""

    unit = "lx"


class DPTColorTemperature(DPT2ByteUnsigned):
    """DPT 7.600 DPT_Color_Temperature (K)."""

    unit = "K"



"""Module for XKNX Exception handling."""
# flake8: noqa
from .exception import XKNXException, ConversionError, \
    CouldNotParseTelegram, CouldNotParseKNXIP, CouldNotParseAddress, \
    DeviceIllegalValue



"""Abstraction to send ConnectRequest and wait for ConnectResponse."""
from xknx.knxip import (
    HPAI, ConnectRequestType, ConnectResponse, KNXIPFrame, KNXIPServiceType)

from .request_response import RequestResponse


class Connect(RequestResponse):
    """Class to send a ConnectRequest and wait for ConnectResponse.."""

    def __init__(self, xknx, udp_client):
        """Initialize Connect class."""
        self.xknx = xknx
        self.udp_client = udp_client
        super(Connect, self).__init__(xknx, self.udp_client, ConnectResponse)
        self.communication_channel = 0
        self.identifier = 0

    def create_knxipframe(self):
        """Create KNX/IP Frame object to be sent to device."""
        (local_addr, local_port) = self.udp_client.getsockname()
        knxipframe = KNXIPFrame(self.xknx)
        knxipframe.init(KNXIPServiceType.CONNECT_REQUEST)
        knxipframe.body.request_type = ConnectRequestType.TUNNEL_CONNECTION

        # set control_endpoint and data_endpoint to the same udp_connection
        knxipframe.body.control_endpoint = HPAI(
            ip_addr=local_addr, port=local_port)
        knxipframe.body.data_endpoint = HPAI(
            ip_addr=local_addr, port=local_port)
        return knxipframe

    def on_success_hook(self, knxipframe):
        """Set communication channel and identifier after having received a valid answer."""
        self.communication_channel = knxipframe.body.communication_channel
        self.identifier = knxipframe.body.identifier
        # Use the address they gave us
        self.xknx.own_address.raw = self.identifier



"""
This package contains all objects managing Tunneling and Routing Connections..

- KNXIPInterface is the overall managing class.
- GatewayScanner searches for available KNX/IP devices in the local network.
- Routing uses UDP/Multicast to communicate with KNX/IP device.
- Tunelling uses UDP packets and builds a static TUnnel with KNX/IP device.
"""
# flake8: noqa
from .request_response import RequestResponse
from .knxip_interface import KNXIPInterface, ConnectionType, ConnectionConfig
from .gateway_scanner import GatewayScanner, GatewayScanFilter
from .routing import Routing
from .tunnel import Tunnel
from .disconnect import Disconnect
from .connectionstate import ConnectionState
from .connect import Connect
from .tunnelling import Tunnelling
from .const import DEFAULT_MCAST_GRP, DEFAULT_MCAST_PORT
from .udp_client import UDPClient



"""Module for reading the values of all devices from device vector from KNX bus in periodic cycles."""
import asyncio


class StateUpdater():
    """Class for reading the values of all devices from KNX bus."""

    def __init__(self,
                 xknx,
                 timeout=3600,
                 start_timeout=10):
        """Initialize StateUpdater class."""
        self.xknx = xknx
        self.timeout = timeout
        self.start_timeout = start_timeout
        self.run_forever = True
        self.run_task = None

    async def start(self):
        """Start StateUpdater."""
        self.run_task = self.xknx.loop.create_task(
            self.run())

    async def run(self):
        """Worker thread. Endless loop for updating states."""
        await asyncio.sleep(self.start_timeout)
        self.xknx.logger.debug("Starting StateUpdater")
        while True:
            await self.xknx.devices.sync()
            await asyncio.sleep(self.timeout)
            if not self.run_forever:
                break



"""Implementation of the KNX date data point."""

import time

from xknx.exceptions import ConversionError

from .dpt import DPTBase


class DPTDate(DPTBase):
    """Abstraction for KNX 3 octet date (DPT 11.001)."""

    @classmethod
    def from_knx(cls, raw):
        """Parse/deserialize from KNX/IP raw data."""
        cls.test_bytesarray(raw, 3)

        day = raw[0] & 0x1F
        month = raw[1] & 0x0F
        year = raw[2] & 0x7F

        if not DPTDate._test_range(day, month, year):
            raise ConversionError("Cant parse DPTDate", raw=raw)

        if year >= 90:
            year += 1900
        else:
            year += 2000

        return {
            'day': day,
            'month': month,
            'year': year
        }

    @classmethod
    def to_knx(cls, values):
        """Serialize to KNX/IP raw data from dict with elements day,month,year."""
        def _knx_year(year):
            if 2000 <= year < 2090:
                return year-2000
            if 1990 <= year < 2000:
                return year-1900
            raise ConversionError("Cant serialize DPTDate", year=year)

        if not isinstance(values, dict):
            raise ConversionError("Cant serialize DPTDate", values=values)
        day = values.get('day', 0)
        month = values.get('month', 0)
        year = _knx_year(values.get('year', 0))

        if not DPTDate._test_range(day, month, year):
            raise ConversionError("Cant serialize DPTDate", values=values)

        return day, month, year

    @classmethod
    def _current_date(cls):
        """Return current local date as struct."""
        localtime = time.localtime()
        day = localtime.tm_mday
        month = localtime.tm_mon
        year = localtime.tm_year
        return {
            'day': day,
            'month': month,
            'year': year
        }

    @classmethod
    def current_date_as_knx(cls):
        """Return current local date as KNX bytes."""
        return cls.to_knx(cls._current_date())

    @staticmethod
    def _test_range(day, month, year):
        """Test if the values are in the correct range."""
        if day < 1 or day > 31:
            return False
        if month < 1 or month > 12:
            return False
        if year < 0 or year > 99:
            return False
        return True



"""
Implementation of KNX Float-values.

They can be either 2 or 4 bytes, and correspond to the the following KDN DPTs.
    9.yyy  2-byte/octet float, e.g. temperature
    14.yyy 4-byte/octet float, IEEE 754, i.e. Electrical measurements: current, power
"""

import struct

from xknx.exceptions import ConversionError

from .dpt import DPTBase


class DPT2ByteFloat(DPTBase):
    """
    Abstraction for KNX 2 Octet Floating Point Numbers.

    DPT 9.xxx
    """

    value_min = -671088.64
    value_max = 670760.96
    unit = ""
    resolution = 1
    payload_length = 2

    @classmethod
    def from_knx(cls, raw):
        """Parse/deserialize from KNX/IP raw data."""
        cls.test_bytesarray(raw, 2)
        data = (raw[0] * 256) + raw[1]
        exponent = (data >> 11) & 0x0f
        significand = data & 0x7ff
        sign = data >> 15

        if sign == 1:
            significand = significand - 2048

        value = float(significand << exponent) / 100

        if not cls._test_boundaries(value):
            raise ConversionError("Cant parse DPT2ByteFloat", value=value)

        return value

    @classmethod
    def to_knx(cls, value):
        """Serialize to KNX/IP raw data."""
        if not isinstance(value, (int, float)):
            raise ConversionError("Cant serialize DPT2ByteFloat", value=value, type=type(value))
        if not cls._test_boundaries(value):
            raise ConversionError("Cant serialize DPT2ByteFloat", value=value)
        sign = 1 if value < 0 else 0

        def calc_exponent(value, sign):
            """Return float exponent."""
            exponent = 0
            significand = abs(int(value * 100))

            while significand < -2048 or significand > 2048:
                exponent += 1
                significand >>= 1

            if sign:
                significand ^= 0x7ff  # invert
                significand += 1     # and add 1

            return exponent, significand

        exponent, significand = calc_exponent(value, sign)

        return (sign << 7) | (exponent << 3) | (significand >> 8), \
            significand & 0xff

    @classmethod
    def _test_boundaries(cls, value):
        """Test if value is within defined range for this object."""
        return cls.value_min <= value <= cls.value_max


class DPT4ByteFloat(DPTBase):
    """
    Abstraction for KNX 4 Octet Floating Point Numbers, with a maximum usable range as specified in IEEE 754.

    The largest positive finite float literal is 3.40282347e+38f.
    The smallest positive finite non-zero literal of type float is 1.40239846e-45f.
    The negative minimum finite float literal is -3.40282347e+38f.
    No value range are defined for DPTs 14.000-079.

    DPT 14.xxx
    """

    unit = ""
    payload_length = 4

    @classmethod
    def from_knx(cls, raw):
        """Parse/deserialize from KNX/IP raw data (big endian)."""
        cls.test_bytesarray(raw, 4)
        try:
            return struct.unpack(">f", bytes(raw))[0]
        except struct.error:
            raise ConversionError("Cant parse DPT4ByteFloat", raw=raw)

    @classmethod
    def to_knx(cls, value):
        """Serialize to KNX/IP raw data."""
        try:
            return tuple(struct.pack(">f", value))
        except struct.error:
            raise ConversionError("Cant serialize DPT4ByteFloat", vlaue=value)


class DPTVoltage(DPT2ByteFloat):
    """
    Abstraction for KNX 2 Octet Floating Point Numbers.

    DPT 9.020 DPT_Value_Voltage.
    """

    unit = "mV"


class DPTPartsPerMillion(DPT2ByteFloat):
    """
    Abstraction for KNX 2 Octet Floating Point Numbers.

    DPT 9.008 DPT_Value_parts/million.
    """

    unit = "ppm"


class DPTEnthalpy(DPT2ByteFloat):
    """
    Abstraction for KNX 2 Octet Floating Point Numbers.

    DPT 9.* 2-byte float value (with unit)
    """

    unit = "H"


class DPTTemperature(DPT2ByteFloat):
    """
    Abstraction for KNX 2 Octet Floating Point Numbers.

    DPT 9.001 DPT_Value_Temp
    """

    value_min = -273
    value_max = 670760
    unit = "°C"
    resolution = 1


class DPTLux(DPT2ByteFloat):
    """
    Abstraction for KNX 2 Octet Floating Point Numbers.

    DPT 9.004 DPT_Value_Lux
    """

    value_min = 0
    value_max = 670760
    unit = "lx"
    resolution = 1


class DPTWsp(DPT2ByteFloat):
    """
    Abstraction for KNX 2 Octet Floating Point Numbers.

    DPT 9.005 DPT_Value_Ws Speed (m/s)
    """

    value_min = 0
    value_max = 670760
    unit = "m/s"
    resolution = 1


class DPTHumidity(DPT2ByteFloat):
    """
    Abstraction for KNX 2 Octet Floating Point Numbers.

    DPT 9.007 DPT_Value_Humidity.
    """

    value_min = 0
    value_max = 670760
    unit = "%"
    resolution = 1


class DPTElectricCurrent(DPT4ByteFloat):
    """DPT 14.019 DPT_Value_Electric_Current."""

    unit = "A"


class DPTElectricPotential(DPT4ByteFloat):
    """DPT 14.027 DPT_Value_Electric_Potential."""

    unit = "V"


class DPTEnergy(DPT4ByteFloat):
    """DPT 14.031 DPT_Value_Energy."""

    unit = 'J'


class DPTFrequency(DPT4ByteFloat):
    """DPT 14.033 DPT_Value_Frequency."""

    unit = 'Hz'


class DPTHeatFlowRate(DPT4ByteFloat):
    """DPT 14.036 DPT_Value_Heat_Flow_Rate."""

    unit = 'W'


class DPTPhaseAngleRad(DPT4ByteFloat):
    """DPT 14.054 DPT_Value_Phase_Angle, Radiant."""

    unit = 'rad'


class DPTPhaseAngleDeg(DPT4ByteFloat):
    """14.055 DPT_Value_Phase_Angle, Degree."""

    unit = '°'


class DPTPower(DPT4ByteFloat):
    """DPT 14.056 DPT_Value_Power."""

    unit = "W"


class DPTPowerFactor(DPT4ByteFloat):
    """DPT 14.057 DPT_Value_Power."""

    unit = ''


class DPTSpeed(DPT4ByteFloat):
    """DPT 14.065 DPT_Value_Speed."""

    unit = 'm/s'



"""
Module for Serialization and Deserialization of a KNX Tunnel ACK information.

Connect requests are used to transmit a KNX telegram within an existing KNX tunnel connection.
With an Tunnel ACK the receiving party acknowledges the valid processing of the request.
"""
from xknx.exceptions import CouldNotParseKNXIP

from .body import KNXIPBody
from .error_code import ErrorCode
from .knxip_enum import KNXIPServiceType


class TunnellingAck(KNXIPBody):
    """Representation of a KNX Connect Request."""

    # pylint: disable=too-many-instance-attributes

    service_type = KNXIPServiceType.TUNNELLING_ACK

    BODY_LENGTH = 4

    def __init__(self, xknx):
        """Initialize TunnellingAck object."""
        super(TunnellingAck, self).__init__(xknx)
        self.communication_channel_id = 1
        self.sequence_counter = 0
        self.status_code = ErrorCode.E_NO_ERROR

    def calculated_length(self):
        """Get length of KNX/IP body."""
        return TunnellingAck.BODY_LENGTH

    def from_knx(self, raw):
        """Parse/deserialize from KNX/IP raw data."""
        def ack_from_knx(header):
            """Parse ACK information."""
            if header[0] != TunnellingAck.BODY_LENGTH:
                raise CouldNotParseKNXIP("connection header wrong length")
            if len(header) < TunnellingAck.BODY_LENGTH:
                raise CouldNotParseKNXIP("connection header wrong length")
            self.communication_channel_id = header[1]
            self.sequence_counter = header[2]
            self.status_code = ErrorCode(header[3])
            return 4
        pos = ack_from_knx(raw)
        return pos

    def to_knx(self):
        """Serialize to KNX/IP raw data."""
        def ack_to_knx():
            """Serialize ACK information."""
            ack = []
            ack.append(TunnellingAck.BODY_LENGTH)
            ack.append(self.communication_channel_id)
            ack.append(self.sequence_counter)
            ack.append(self.status_code.value)
            return ack
        data = []
        data.extend(ack_to_knx())
        return data

    def __str__(self):
        """Return object as readable string."""
        return '<TunnellingAck communication_channel_id="{0}" ' \
            'sequence_counter="{1}" status_code="{2}" />' \
            .format(self.communication_channel_id, self.sequence_counter,
                    self.status_code)



"""
Module for managing a DTP Scene Number remote value.

DPT 17.001.
"""
from xknx.knx import DPTArray, DPTSceneNumber

from .remote_value import RemoteValue


class RemoteValueSceneNumber(RemoteValue):
    """Abstraction for remote value of KNX DPT 17.001 (DPT_Scene_Number)."""

    def __init__(self,
                 xknx,
                 group_address=None,
                 device_name=None,
                 after_update_cb=None):
        """Initialize remote value of KNX DPT 17.001 (DPT_Scene_Number)."""
        # pylint: disable=too-many-arguments
        super(RemoteValueSceneNumber, self).__init__(
            xknx, group_address, None,
            device_name=device_name, after_update_cb=after_update_cb)

    def payload_valid(self, payload):
        """Test if telegram payload may be parsed."""
        return (isinstance(payload, DPTArray)
                and len(payload.value) == 1)

    def to_knx(self, value):
        """Convert value to payload."""
        return DPTArray(DPTSceneNumber.to_knx(value))

    def from_knx(self, payload):
        """Convert current payload to value."""
        return DPTSceneNumber.from_knx(payload.value)



"""
Module for KNX Telegrams.

The telegram class is the leightweight interaction object between

* business logic (Lights, Covers, etc) and
* underlaying KNX/IP abstraction (KNX-Routing/KNX-Tunneling).

It contains

* the telegram type (e.g. GROUP_WRITE)
* the direction (incoming or outgoing)
* the group address (e.g. 1/2/3)
* and the payload (e.g. "12%" or "23.23 C".

"""
from enum import Enum

from .address import GroupAddress


class TelegramDirection(Enum):
    """Enum class for the communication direction of a telegram (from KNX bus or to KNX bus)."""

    INCOMING = 1
    OUTGOING = 2


class TelegramType(Enum):
    """Enum class for type of telegram."""

    GROUP_READ = 1
    GROUP_WRITE = 2
    GROUP_RESPONSE = 3


class Telegram:
    """Class for KNX telegrams."""

    # pylint: disable=too-few-public-methods

    def __init__(self, group_address=GroupAddress(None),
                 telegramtype=TelegramType.GROUP_WRITE,
                 direction=TelegramDirection.OUTGOING,
                 payload=None):
        """Initialize Telegram class."""
        self.direction = direction
        self.telegramtype = telegramtype
        self.group_address = group_address
        self.payload = payload

    def __str__(self):
        """Return object as readable string."""
        return '<Telegram group_address="{0}", payload="{1}" ' \
            'telegramtype="{2}" direction="{3}" />'.format(
                self.group_address.__repr__(),
                self.payload,
                self.telegramtype,
                self.direction)

    def __eq__(self, other):
        """Equal operator."""
        return self.__dict__ == other.__dict__



"""Implementation of different KNX DPT HVAC Operation modes."""

from enum import Enum

from xknx.exceptions import ConversionError, CouldNotParseKNXIP

from .dpt import DPTBase


class HVACOperationMode(Enum):
    """Enum for the different KNX HVAC operation modes."""

    AUTO = "Auto"
    COMFORT = "Comfort"
    STANDBY = "Standby"
    NIGHT = "Night"
    FROST_PROTECTION = "Frost Protection"
    HEAT = "Heat"
    MORNING_WARMUP = "Morning Warmup"
    COOL = "Cool"
    NIGHT_PURGE = "Night Purge"
    PRECOOL = "Precool"
    OFF = "Off"
    TEST = "Test"
    EMERGENCY_HEAT = "Emergency Heat"
    FAN_ONLY = "Fan only"
    FREE_COOL = "Free Cool"
    ICE = "Ice"
    DRY = "Dry"
    NODEM = "NoDem"


class DPTHVACMode(DPTBase):
    """
    Abstraction for KNX KNX HVAC mod.

    DPT 20.102
    """

    @classmethod
    def from_knx(cls, raw):
        """Parse/deserialize from KNX/IP raw data."""
        cls.test_bytesarray(raw, 1)
        if raw[0] == 0x04:
            return HVACOperationMode.FROST_PROTECTION
        if raw[0] == 0x03:
            return HVACOperationMode.NIGHT
        if raw[0] == 0x02:
            return HVACOperationMode.STANDBY
        if raw[0] == 0x01:
            return HVACOperationMode.COMFORT
        if raw[0] == 0x00:
            return HVACOperationMode.AUTO
        raise CouldNotParseKNXIP("Could not parse HVACOperationMode")

    @classmethod
    def to_knx(cls, value):
        """Serialize to KNX/IP raw data."""
        if value == HVACOperationMode.AUTO:
            return (0,)
        if value == HVACOperationMode.COMFORT:
            return (1,)
        if value == HVACOperationMode.STANDBY:
            return (2,)
        if value == HVACOperationMode.NIGHT:
            return (3,)
        if value == HVACOperationMode.FROST_PROTECTION:
            return (4,)
        raise ConversionError("Could not parse HVACOperationMode", value=value)


class DPTControllerStatus(DPTBase):
    """
    Abstraction for KNX HVAC Controller status.

    Non-standardised DP type (in accordance with KNX AN 097/07 rev 3).

    Help needed:
    The values of this type were retrieved by reverse engineering. Any
    notes on the correct implementation of this type are highly appreciated.
    """

    @classmethod
    def from_knx(cls, raw):
        """Parse/deserialize from KNX/IP raw data."""
        cls.test_bytesarray(raw, 1)
        if raw[0] & 8 > 0:
            return HVACOperationMode.FROST_PROTECTION
        if raw[0] & 4 > 0:
            return HVACOperationMode.NIGHT
        if raw[0] & 2 > 0:
            return HVACOperationMode.STANDBY
        if raw[0] & 1 > 0:
            return HVACOperationMode.COMFORT
        raise CouldNotParseKNXIP("Could not parse HVACOperationMode")

    @classmethod
    def to_knx(cls, value):
        """Serialize to KNX/IP raw data."""
        if value == HVACOperationMode.AUTO:
            raise ConversionError("Cant serialize DPTControllerStatus", value=value)
        elif value == HVACOperationMode.COMFORT:
            return (0x21,)
        elif value == HVACOperationMode.STANDBY:
            return (0x22,)
        elif value == HVACOperationMode.NIGHT:
            return (0x24,)
        elif value == HVACOperationMode.FROST_PROTECTION:
            return (0x28,)
        raise ConversionError("Could not parse HVACOperationMode", value=value)



"""Implementation of 3.17 Datapoint Types String."""
from xknx.exceptions import ConversionError

from .dpt import DPTBase


class DPTString(DPTBase):
    """
    Abstraction for KNX 14 Octet String.

    DPT 3.17
    """

    STRING_SIZE = 14

    @classmethod
    def from_knx(cls, raw):
        """Parse/deserialize from KNX/IP raw data."""
        cls.test_bytesarray(raw, cls.STRING_SIZE)
        value = str()
        for byte in raw:
            if byte != 0x00:
                value += chr(byte)
        return value

    @classmethod
    def to_knx(cls, value):
        """Serialize to KNX/IP raw data."""
        if not isinstance(value, str):
            raise ConversionError("Cant serialize DPTString", value=value)
        if not cls._test_boundaries(value):
            raise ConversionError("Cant serialize DPTString", value=value)

        raw = []
        for character in value:
            raw.append(ord(character))
        raw.extend([0] * (cls.STRING_SIZE - len(raw)))
        return raw

    @classmethod
    def _test_boundaries(cls, value):
        """Test if value is within defined range for this object."""
        return len(value) <= cls.STRING_SIZE



"""Module for handling commands which may be attached to BinarySensor class."""


class ActionBase():
    """Base Class for handling commands."""

    def __init__(self,
                 xknx,
                 hook="on",
                 counter=1):
        """Initialize Action_Base class."""
        self.xknx = xknx
        self.hook = hook
        self.counter = counter

    def test_counter(self, counter):
        """Test if action filters for specific counter."""
        if self.counter is None:
            # no specific counter_filter -> always true
            return True
        if counter is None:
            return True
        return counter == self.counter

    def test_if_applicable(self, state, counter=None):
        """Test if should be executed for this state and this counter number."""
        from .binary_sensor import BinarySensorState
        if (state == BinarySensorState.ON) \
                and (self.hook == "on") \
                and self.test_counter(counter):
            return True
        if (state == BinarySensorState.OFF) \
                and (self.hook == "off") \
                and self.test_counter(counter):
            return True
        return False

    async def execute(self):
        """Execute action. To be overwritten in derived classes."""
        self.xknx.logger.info("Execute not implemented for %s", self.__class__.__name__)

    def __str__(self):
        """Return object as readable string."""
        return '<ActionBase hook="{0}" counter="{1}"/>' \
            .format(self.hook, self.counter)

    def __eq__(self, other):
        """Equal operator."""
        return self.__dict__ == other.__dict__


class Action(ActionBase):
    """Class for handling commands."""

    def __init__(self,
                 xknx,
                 hook="on",
                 target=None,
                 method=None,
                 counter=1):
        """Initialize Action class."""
        # pylint: disable=too-many-arguments
        super(Action, self).__init__(xknx, hook, counter)
        self.target = target
        self.method = method

    @classmethod
    def from_config(cls, xknx, config):
        """Initialize object from configuration structure."""
        hook = config.get("hook", "on")
        target = config.get("target")
        method = config.get("method")
        counter = config.get("counter", 1)
        return cls(xknx,
                   hook=hook,
                   target=target,
                   method=method,
                   counter=counter)

    async def execute(self):
        """Execute action."""
        if self.target is not None:
            if self.target not in self.xknx.devices:
                self.xknx.logger.warning("Unknown device %s witin action %s.", self.target, self)
                return
            await self.xknx.devices[self.target].do(self.method)

    def __str__(self):
        """Return object as readable string."""
        return '<Action target="{0}" method="{1}" {2}/>' \
            .format(self.target, self.method, super(Action, self).__str__())


class ActionCallback(ActionBase):
    """Class for handling commands via callbacks."""

    def __init__(self,
                 xknx,
                 callback,
                 hook="on",
                 counter=1):
        """Initialize Action class."""
        # pylint: disable=too-many-arguments
        super(ActionCallback, self).__init__(xknx, hook, counter)
        self.callback = callback

    async def execute(self):
        """Execute callback."""
        await self.callback()

    def __str__(self):
        """Return object as readable string."""
        return '<ActionCallback callback="{}" {}/>' \
            .format(self.callback.__name__, super(ActionCallback, self).__str__())



"""
Module for managing a DTP 5010 remote value.

DPT 5.010.
"""
from xknx.knx import DPTArray, DPTValue1Ucount

from .remote_value import RemoteValue


class RemoteValueDptValue1Ucount(RemoteValue):
    """Abstraction for remote value of KNX DPT 5.010."""

    def __init__(self,
                 xknx,
                 group_address=None,
                 device_name=None,
                 after_update_cb=None):
        """Initialize remote value of KNX DPT 5.010."""
        # pylint: disable=too-many-arguments
        super(RemoteValueDptValue1Ucount, self).__init__(
            xknx, group_address, None,
            device_name=device_name, after_update_cb=after_update_cb)

    def payload_valid(self, payload):
        """Test if telegram payload may be parsed."""
        return (isinstance(payload, DPTArray)
                and len(payload.value) == 1)

    def to_knx(self, value):
        """Convert value to payload."""
        return DPTArray(DPTValue1Ucount.to_knx(value))

    def from_knx(self, payload):
        """Convert current payload to value."""
        return DPTValue1Ucount.from_knx(payload.value)



"""Setup for XKNX python package."""
from setuptools import find_packages, setup

VERSION = '0.9.4'

REQUIRES = [
    'pyyaml>=3.12',
    'netifaces>=0.10.0',
    ]

setup(
    name='xknx',
    description='An Asynchronous Library for the KNX protocol. Documentation: http://xknx.io/',
    version=VERSION,
    download_url='https://github.com/XKNX/xknx/archive/{}.zip'.format(VERSION),
    url='http://xknx.io/',
    author='Julius Mittenzwei',
    author_email='julius@mittenzwei.com',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'Topic :: System :: Hardware :: Hardware Drivers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ],
    packages=find_packages(),
    install_requires=REQUIRES,
    # python_requires=">=3.5.2",
    keywords='knx ip knxip eib home automation',
    zip_safe=False)



"""
Module for serialization and deserialization of KNX DIB information.

DIB is Description Information Block.

A KNX/IP Search Response may contain several DIBs of different types:

* DIBSuppSVCFamilies:   Supported features of device
* DIBDeviceInformation: Name, serial number, some unimportant flags
* DIBGeneric:           General Information
                        (fallback for unknown dib type codes)
"""

from xknx.exceptions import CouldNotParseKNXIP
from xknx.knx import PhysicalAddress

from .knxip_enum import DIBServiceFamily, DIBTypeCode, KNXMedium


class DIB():
    """
    Base class for DIB (Description Information Block).

    This base class is only the interface for the derived
    classes.
    """

    def __init__(self):
        """Initialize DIB class."""

    def calculated_length(self):
        """Get length of KNX/IP object."""

    def from_knx(self, raw):
        """Parse/deserialize from KNX/IP raw data."""

    def to_knx(self):
        """Serialize to KNX/IP raw data."""

    @staticmethod
    def determine_dib(raw):
        """Determine dib type out of dib type code."""
        if len(raw) < 2:
            raise CouldNotParseKNXIP("could not parse DIB header")
        dtc = DIBTypeCode(raw[1])

        if dtc == DIBTypeCode.DEVICE_INFO:
            return DIBDeviceInformation()
        if dtc == DIBTypeCode.SUPP_SVC_FAMILIES:
            return DIBSuppSVCFamilies()
        return DIBGeneric()


class DIBGeneric(DIB):
    """
    Module for serialization and deserialization of KNX DIB Generic.

    Fallback for not implemented DIBTypeCodes.
    """

    def __init__(self):
        """Initialize DIBGeneric class."""
        super(DIBGeneric, self).__init__()
        # DTC Description Type Code
        self.dtc = None
        # IBD Information Block Data
        self.data = []

    def calculated_length(self):
        """Get length of KNX/IP object."""
        return len(self.data)

    def from_knx(self, raw):
        """Parse/deserialize from KNX/IP raw data."""
        if len(raw) < 2:
            raise CouldNotParseKNXIP("could not parse DIB header")

        dib_length = raw[0]
        if len(raw) < dib_length:
            raise CouldNotParseKNXIP("DIB wrong length")

        self.dtc = DIBTypeCode(raw[1])
        self.data = raw[:dib_length]

        return dib_length

    def to_knx(self):
        """Serialize to KNX/IP raw data."""
        data = []
        data.append(len(self.data))
        data.append(self.dtc.value)
        data.extend(self.data[2:])
        return data

    def __str__(self):
        """Return object as readable string."""
        return '<DIB dtc="{0}" data="{1}" />'.format(
            self.dtc,
            ', '.join('0x%02x' % i for i in self.data))


class DIBSuppSVCFamilies(DIB):
    """Class for serialization and deserialization of KNX DIB Supported Services."""

    # pylint: disable=too-few-public-methods

    class Family:
        """Class for storing a supported device family."""

        def __init__(self, name=None, version=None):
            """Initialize DIBSuppSVCFamilies.Family."""
            self.name = name
            self.version = version

        def __str__(self):
            """Return object as readable string."""
            return '<Family name="{0}" version="{1}" />' \
                .format(self.name, self.version)

        def __eq__(self, other):
            """Equal operator."""
            return self.__dict__ == other.__dict__

    def __init__(self):
        """Initialize DIBSuppSVCFamilies class."""
        super(DIBSuppSVCFamilies, self).__init__()
        self.families = []

    def supports(self, name):
        """Return if device supports a given service family by name."""
        for family in self.families:
            if name == family.name:
                return True
        return False

    def calculated_length(self):
        """Get length of KNX/IP object."""
        return len(self.families)*2+2

    def from_knx(self, raw):
        """Parse/deserialize from KNX/IP raw data."""
        if len(raw) < 2:
            raise CouldNotParseKNXIP("DIB header too small")
        length = raw[0]
        if len(raw) < length:
            raise CouldNotParseKNXIP("DIB wrong size")
        if DIBTypeCode(raw[1]) != DIBTypeCode.SUPP_SVC_FAMILIES:
            raise CouldNotParseKNXIP("DIB is no device info")

        for i in range(0, int((length-2)/2)):
            name = DIBServiceFamily(raw[i*2+2])
            version = raw[i*2+3]
            self.families.append(DIBSuppSVCFamilies.Family(name, version))
        return length

    def to_knx(self):
        """Serialize to KNX/IP raw data."""
        data = []
        data.append(len(self.families)*2+2)
        data.append(DIBTypeCode.SUPP_SVC_FAMILIES.value)
        for family in self.families:
            data.append(family.name.value)
            data.append(family.version)
        return data

    def __str__(self):
        """Return object as readable string."""
        return '<DIBSuppSVCFamilies families="[{0}]" />' \
            .format(", ".join("{0} version: {1}".format(
                family.name, family.version) for family in self.families))


class DIBDeviceInformation(DIB):
    """Class for serialization and deserialization of KNX DIB Device Information Block."""

    # pylint: disable=too-many-instance-attributes

    LENGTH = 54

    def __init__(self):
        """Initialize DIBDeviceInformation class."""
        super(DIBDeviceInformation, self).__init__()
        self.knx_medium = KNXMedium.TP1
        self.programming_mode = False
        self.individual_address = PhysicalAddress(None)
        self.installation_number = 0
        self.project_number = 0
        self.serial_number = ""
        self.multicast_address = "224.0.23.12"
        self.mac_address = ""
        self.name = ""

    def calculated_length(self):
        """Get length of KNX/IP object."""
        return DIBDeviceInformation.LENGTH

    def from_knx(self, raw):
        """Parse/deserialize from KNX/IP raw data."""
        if len(raw) < DIBDeviceInformation.LENGTH:
            raise CouldNotParseKNXIP("wrong connection header length")
        if raw[0] != DIBDeviceInformation.LENGTH:
            raise CouldNotParseKNXIP("wrong connection header length")
        if DIBTypeCode(raw[1]) != DIBTypeCode.DEVICE_INFO:
            raise CouldNotParseKNXIP("DIB is no device info")

        self.knx_medium = KNXMedium(raw[2])
        # last bit of device_status. All other bits are unused
        self.programming_mode = bool(raw[3])
        self.individual_address = \
            PhysicalAddress((raw[4], raw[5]))
        installation_project_identifier = raw[6]*256+raw[7]
        self.project_number = installation_project_identifier >> 4
        self.installation_number = installation_project_identifier & 15
        self.serial_number = ":".join('%02x' % i for i in raw[8:14])
        self.multicast_address = ".".join('%i' % i for i in raw[14:18])
        self.mac_address = ":".join('%02x' % i for i in raw[18:24])
        self.name = "".join(map(chr, raw[24:54])).rstrip('\0')
        return DIBDeviceInformation.LENGTH

    def to_knx(self):
        """Serialize to KNX/IP raw data."""
        def hex_notation_to_knx(serial_number):
            """Serialize hex notation."""
            for part in serial_number.split(":"):
                yield int(part, 16)

        def ip_to_knx(ip_addr):
            """Serialize ip."""
            for part in ip_addr.split("."):
                yield int(part)

        def str_to_knx(string, length):
            """Serialize string."""
            if len(string) > length-1:
                string = string[:length-1]
            for char in string:
                yield ord(char)
            for _ in range(0, 30-len(string)):
                yield 0x00
        installation_project_identifier = \
            (self.project_number * 16) + \
            self.installation_number
        data = []
        data.append(DIBDeviceInformation.LENGTH)
        data.append(DIBTypeCode.DEVICE_INFO.value)
        data.append(self.knx_medium.value)
        data.append(int(self.programming_mode))
        data.extend(self.individual_address.to_knx())
        data.append((installation_project_identifier >> 8) & 255)
        data.append(installation_project_identifier & 255)
        data.extend(hex_notation_to_knx(self.serial_number))
        data.extend(ip_to_knx(self.multicast_address))
        data.extend(hex_notation_to_knx(self.mac_address))
        data.extend(str_to_knx(self.name, 30))
        return data

    def __str__(self):
        """Return object as readable string."""
        return '<DIBDeviceInformation ' \
               '\n\tknx_medium="{0}" ' \
               '\n\tprogramming_mode="{1}" ' \
               '\n\tindividual_address="{2}" ' \
               '\n\tinstallation_number="{3}" ' \
               '\n\tproject_number="{4}" ' \
               '\n\tserial_number="{5}" ' \
               '\n\tmulticast_address="{6}" ' \
               '\n\tmac_address="{7}" ' \
               '\n\tname="{8}" />'.format(
                   self.knx_medium, self.programming_mode,
                   self.individual_address, self.installation_number,
                   self.project_number, self.serial_number,
                   self.multicast_address, self.mac_address, self.name)



"""
Module for Serialization and Deserialization of a KNX Connect Response information.

Connect requests are used to start a new tunnel connection on a KNX/IP device.
With an Connect Response the receiving party acknowledges the valid processing of the request.
"""
from xknx.exceptions import CouldNotParseKNXIP

from .body import KNXIPBody
from .error_code import ErrorCode
from .hpai import HPAI
from .knxip_enum import ConnectRequestType, KNXIPServiceType


class ConnectResponse(KNXIPBody):
    """Representation of a KNX Connect Response."""

    # pylint: disable=too-many-instance-attributes

    service_type = KNXIPServiceType.CONNECT_RESPONSE

    CRD_LENGTH = 4

    def __init__(self, xknx):
        """Initialize ConnectResponse class."""
        super(ConnectResponse, self).__init__(xknx)

        self.communication_channel = 0
        self.status_code = ErrorCode.E_NO_ERROR
        self.request_type = None
        self.control_endpoint = HPAI()
        self.identifier = None

    def calculated_length(self):
        """Get length of KNX/IP body."""
        return 2 + HPAI.LENGTH + \
            ConnectResponse.CRD_LENGTH

    def from_knx(self, raw):
        """Parse/deserialize from KNX/IP raw data."""
        def crd_from_knx(crd):
            """Parse CRD (Connection Response Data Block)."""
            if crd[0] != ConnectResponse.CRD_LENGTH:
                raise CouldNotParseKNXIP("CRD has wrong length")
            if len(crd) < ConnectResponse.CRD_LENGTH:
                raise CouldNotParseKNXIP("CRD data has wrong length")
            self.request_type = ConnectRequestType(crd[1])
            self.identifier = crd[2]*256+crd[3]
            return 4

        self.communication_channel = raw[0]
        self.status_code = ErrorCode(raw[1])
        pos = 2

        pos += self.control_endpoint.from_knx(raw[pos:])
        pos += crd_from_knx(raw[pos:])
        return pos

    def to_knx(self):
        """Serialize to KNX/IP raw data."""
        def crd_to_knx():
            """Serialize CRD (Connect Response Data Block)."""
            crd = []
            crd.append(ConnectResponse.CRD_LENGTH)
            crd.append(self.request_type.value)
            crd.append((self.identifier >> 8) & 255)
            crd.append(self.identifier & 255)
            return crd
        data = []
        data.append(self.communication_channel)
        data.append(self.status_code.value)
        data.extend(self.control_endpoint.to_knx())
        data.extend(crd_to_knx())
        return data

    def __str__(self):
        """Return object as readable string."""
        return '<ConnectResponse communication_channel="{0}" ' \
            'status_code="{1}" control_endpoint="{2}" ' \
            'request_type="{3}" identifier="{4}" />' \
            .format(self.communication_channel, self.status_code,
                    self.control_endpoint, self.request_type,
                    self.identifier)



"""
Module for managing a sensor via KNX.

It provides functionality for

* reading the current state from KNX bus.
* watching for state updates from KNX bus.
"""
from .device import Device
from .remote_value_dpt_value_1_ucount import RemoteValueDptValue1Ucount
from .remote_value_scaling import RemoteValueScaling
from .remote_value_sensor import RemoteValueSensor


class Sensor(Device):
    """Class for managing a sensor."""

    def __init__(self,
                 xknx,
                 name,
                 group_address=None,
                 value_type=None,
                 device_updated_cb=None):
        """Initialize Sensor class."""
        # pylint: disable=too-many-arguments
        super(Sensor, self).__init__(xknx, name, device_updated_cb)

        self.sensor_value = None
        if value_type == "percent":
            self.sensor_value = RemoteValueScaling(
                xknx,
                group_address_state=group_address,
                device_name=self.name,
                after_update_cb=self.after_update,
                range_from=0,
                range_to=100)
        elif value_type == "pulse":
            self.sensor_value = RemoteValueDptValue1Ucount(
                xknx,
                group_address=group_address,
                device_name=self.name,
                after_update_cb=self.after_update)
        else:
            self.sensor_value = RemoteValueSensor(
                xknx,
                group_address_state=group_address,
                device_name=self.name,
                after_update_cb=self.after_update,
                value_type=value_type)

    @classmethod
    def from_config(cls, xknx, name, config):
        """Initialize object from configuration structure."""
        group_address = \
            config.get('group_address')
        value_type = \
            config.get('value_type')

        return cls(xknx,
                   name,
                   group_address=group_address,
                   value_type=value_type)

    def has_group_address(self, group_address):
        """Test if device has given group address."""
        return self.sensor_value.has_group_address(group_address)

    def state_addresses(self):
        """Return group addresses which should be requested to sync state."""
        return self.sensor_value.state_addresses()

    async def process_group_write(self, telegram):
        """Process incoming GROUP WRITE telegram."""
        await self.sensor_value.process(telegram)

    def unit_of_measurement(self):
        """Return the unit of measurement."""
        return self.sensor_value.unit_of_measurement

    def resolve_state(self):
        """Return the current state of the sensor as a human readable string."""
        return self.sensor_value.value

    def __str__(self):
        """Return object as readable string."""
        return '<Sensor name="{0}" ' \
               'sensor="{1}" value="{2}" unit="{3}"/>' \
            .format(self.name,
                    self.sensor_value.group_addr_str(),
                    self.resolve_state(),
                    self.unit_of_measurement())

    def __eq__(self, other):
        """Equal operator."""
        return self.__dict__ == other.__dict__



"""Implementation of Basic KNX Time."""

import time

from xknx.exceptions import ConversionError

from .dpt import DPTBase, DPTWeekday


class DPTTime(DPTBase):
    """
    Abstraction for KNX 3 Octet Time.

    DPT 10.001
    """

    @classmethod
    def from_knx(cls, raw):
        """Parse/deserialize from KNX/IP raw data."""
        cls.test_bytesarray(raw, 3)

        weekday = (raw[0] & 0xE0) >> 5
        hours = raw[0] & 0x1F
        minutes = raw[1] & 0x3F
        seconds = raw[2] & 0x3F

        if not DPTTime._test_range(weekday, hours, minutes, seconds):
            raise ConversionError("Cant parse DPTTime", raw=raw)

        return {'weekday': DPTWeekday(weekday),
                'hours': hours,
                'minutes': minutes,
                'seconds': seconds}

    @classmethod
    def to_knx(cls, values):
        """Serialize to KNX/IP raw data from dict with elements weekday,hours,minutes,seconds."""
        if not isinstance(values, dict):
            raise ConversionError("Cant serialize DPTTime", values=values)
        weekday = values.get('weekday', DPTWeekday.NONE).value
        hours = values.get('hours', 0)
        minutes = values.get('minutes', 0)
        seconds = values.get('seconds', 0)

        if not DPTTime._test_range(weekday, hours, minutes, seconds):
            raise ConversionError("Cant serialize DPTTime", values=values)

        return weekday << 5 | hours, minutes, seconds

    @classmethod
    def _current_time(cls):
        """Return current local time as struct."""
        localtime = time.localtime()
        weekday = localtime.tm_wday + 1
        hours = localtime.tm_hour
        minutes = localtime.tm_min
        seconds = localtime.tm_sec
        return {'weekday': DPTWeekday(weekday),
                'hours': hours,
                'minutes': minutes,
                'seconds': seconds}

    @classmethod
    def current_time_as_knx(cls):
        """Return current local time as KNX bytes."""
        return cls.to_knx(cls._current_time())

    @staticmethod
    def _test_range(weekday, hours, minutes, seconds):
        """Test if values are in the correct value range."""
        if weekday < 0 or weekday > 7:
            return False
        if hours < 0 or hours > 23:
            return False
        if minutes < 0 or minutes > 59:
            return False
        if seconds < 0 or seconds > 59:
            return False
        return True



"""
Module for managing a remote value typically used within a sensor.

The module maps a given value_type to a DPT class and uses this class
for serialization and deserialization of the KNX value.
"""
from xknx.exceptions import ConversionError
from xknx.knx import (
    DPT2ByteFloat, DPT2ByteUnsigned, DPT4ByteFloat, DPT4ByteSigned,
    DPT4ByteUnsigned, DPTArray, DPTBrightness, DPTColorTemperature,
    DPTElectricCurrent, DPTElectricPotential, DPTEnergy, DPTEnthalpy,
    DPTFrequency, DPTHeatFlowRate, DPTHumidity, DPTLux, DPTPartsPerMillion,
    DPTPhaseAngleDeg, DPTPhaseAngleRad, DPTPower, DPTPowerFactor, DPTSpeed,
    DPTTemperature, DPTUElCurrentmA, DPTVoltage, DPTWsp)

from .remote_value import RemoteValue


class RemoteValueSensor(RemoteValue):
    """Abstraction for many different sensor DPT types."""

    DPTMAP = {
        'current': DPTUElCurrentmA,
        'brightness': DPTBrightness,
        'color_temperature': DPTColorTemperature,
        'temperature': DPTTemperature,
        'illuminance': DPTLux,
        'speed_ms': DPTWsp,
        'humidity': DPTHumidity,
        'voltage': DPTVoltage,
        'power': DPTPower,
        'electric_current': DPTElectricCurrent,
        'electric_potential': DPTElectricPotential,
        'energy': DPTEnergy,
        'frequency': DPTFrequency,
        'heatflowrate': DPTHeatFlowRate,
        'phaseanglerad': DPTPhaseAngleRad,
        'phaseangledeg': DPTPhaseAngleDeg,
        'powerfactor': DPTPowerFactor,
        'speed': DPTSpeed,
        'enthalpy': DPTEnthalpy,
        'ppm': DPTPartsPerMillion,

        #  Generic DPT Without Min/Max and Unit.
        'DPT-7': DPT2ByteUnsigned,
        '2byte_unsigned': DPT2ByteUnsigned,
        'DPT-9': DPT2ByteFloat,

        'DPT-12': DPT4ByteUnsigned,
        '4byte_unsigned': DPT4ByteUnsigned,

        'DPT-13': DPT4ByteSigned,
        '4byte_signed': DPT4ByteSigned,
        'DPT-14': DPT4ByteFloat,
        '4byte_float': DPT4ByteFloat
    }

    def __init__(self,
                 xknx,
                 group_address=None,
                 group_address_state=None,
                 value_type=None,
                 device_name=None,
                 after_update_cb=None):
        """Initialize RemoteValueSensor class."""
        # pylint: disable=too-many-arguments
        super(RemoteValueSensor, self).__init__(
            xknx, group_address, group_address_state,
            device_name=device_name, after_update_cb=after_update_cb)
        if value_type not in self.DPTMAP:
            raise ConversionError("invalid value type", value_type=value_type, device_name=device_name)
        self.value_type = value_type

    def payload_valid(self, payload):
        """Test if telegram payload may be parsed."""
        return (
            isinstance(payload, DPTArray) and
            len(payload.value) == self.DPTMAP[self.value_type].payload_length)

    def to_knx(self, value):
        """Convert value to payload."""
        return DPTArray(self.DPTMAP[self.value_type].to_knx(value))

    def from_knx(self, payload):
        """Convert current payload to value."""
        return self.DPTMAP[self.value_type].from_knx(payload.value)

    @property
    def unit_of_measurement(self):
        """Return the unit of measurement."""
        return self.DPTMAP[self.value_type].unit



"""Implementation of the KNX datetime data point."""

import time

from xknx.exceptions import ConversionError

from .dpt import DPTBase, DPTWeekday


class DPTDateTime(DPTBase):
    """Abstraction for KNX 8 octet datetime (DPT 19.001)."""

    @classmethod
    def from_knx(cls, raw):
        """Parse/deserialize from KNX/IP raw data."""
        cls.test_bytesarray(raw, 8)

        year = raw[0] + 1900
        month = raw[1] & 0x0F
        day = raw[2] & 0x3F
        weekday = (raw[3] & 0xE0) >> 5
        hours = raw[3] & 0x1F
        minutes = raw[4] & 0x3F
        seconds = raw[5] & 0x3F

        if not DPTDateTime._test_range(year, month, day, weekday, hours, minutes, seconds):
            raise ConversionError("Could not parse DPTDateTime", raw=raw)

        return {
            'year': year,
            'month': month,
            'day': day,
            'weekday': DPTWeekday(weekday),
            'hours': hours,
            'minutes': minutes,
            'seconds': seconds
        }

    @classmethod
    def to_knx(cls, values):
        """Serialize to KNX/IP raw data from dict with elements year,month,day,weekday,hours,minutes,seconds."""
        if not isinstance(values, dict):
            raise ConversionError("Cant serialize DPTDateTime", values=values)

        year = values.get('year', 1900)
        month = values.get('month', 1)
        day = values.get('day', 1)
        weekday = values.get('weekday', DPTWeekday.NONE).value
        hours = values.get('hours', 0)
        minutes = values.get('minutes', 0)
        seconds = values.get('seconds', 0)

        if not DPTDateTime._test_range(year, month, day, weekday, hours, minutes, seconds):
            raise ConversionError("Cant serialize DPTDateTime", values=values)

        return year - 1900, month, day, weekday << 5 | hours, minutes, seconds, 0, 0

    @classmethod
    def current_datetime_as_knx(cls):
        """Return current local datetime as KNX bytes."""
        return cls.to_knx(cls._current_datetime())

    @classmethod
    def _current_datetime(cls):
        """Return current local time as struct."""
        localtime = time.localtime()
        year = localtime.tm_year
        month = localtime.tm_mon
        day = localtime.tm_mday
        weekday = localtime.tm_wday + 1
        hours = localtime.tm_hour
        minutes = localtime.tm_min
        seconds = localtime.tm_sec
        return {
            'year': year,
            'month': month,
            'day': day,
            'weekday': DPTWeekday(weekday),
            'hours': hours,
            'minutes': minutes,
            'seconds': seconds
        }

    @staticmethod
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-return-statements
    def _test_range(year, month, day, weekday, hours, minutes, seconds):
        """Test if the values are in the correct range."""
        if year < 1900 or year > 2155:
            return False
        if month < 1 or month > 12:
            return False
        if day < 1 or day > 31:
            return False
        if weekday < 0 or weekday > 7:
            return False
        if hours < 0 or hours > 23:
            return False
        if minutes < 0 or minutes > 59:
            return False
        if seconds < 0 or seconds > 59:
            return False
        return True



"""Module for XKXN Exceptions."""


class XKNXException(Exception):
    """Default XKNX Exception."""

    def __eq__(self, other):
        """Equal operator."""
        return self.__dict__ == other.__dict__

    def __hash__(self):
        """Hash function."""
        return 0


class CouldNotParseTelegram(XKNXException):
    """Could not parse telegram error."""

    def __init__(self, description, **kwargs):
        """Initialize CouldNotParseTelegram class."""
        super(CouldNotParseTelegram, self).__init__("Could not parse Telegram")
        self.description = description
        self.parameter = kwargs

    def _format_parameter(self):
        return " ".join(['%s="%s"' % (key, value) for (key, value) in sorted(self.parameter.items())])

    def __str__(self):
        """Return object as readable string."""
        return '<CouldNotParseTelegram description="{0}" {1}/>' \
            .format(self.description, self._format_parameter())


class CouldNotParseKNXIP(XKNXException):
    """Exception falss for worong KNXIP data."""

    def __init__(self, description=""):
        """Initialize CouldNotParseTelegram class."""
        super(CouldNotParseKNXIP, self).__init__("Could not parse KNXIP")
        self.description = description

    def __str__(self):
        """Return object as readable string."""
        return '<CouldNotParseKNXIP description="{0}" />' \
            .format(self.description)


class ConversionError(XKNXException):
    """Exception class for error while converting one type to another."""

    def __init__(self, description, **kwargs):
        """Initialize ConversionError class."""
        super(ConversionError, self).__init__("Conversion Error")
        self.description = description
        self.parameter = kwargs

    def _format_parameter(self):
        return " ".join(['%s="%s"' % (key, value) for (key, value) in sorted(self.parameter.items())])

    def __str__(self):
        """Return object as readable string."""
        return '<ConversionError description="{0}" {1}/>'.format(self.description, self._format_parameter())


class CouldNotParseAddress(XKNXException):
    """Exception class for wrong address format."""

    def __init__(self, address=None):
        """Initialize CouldNotParseAddress class."""
        super(CouldNotParseAddress, self).__init__("Could not parse address")
        self.address = address

    def __str__(self):
        """Return object as readable string."""
        return '<CouldNotParseAddress address="{0}" />'.format(self.address)


class DeviceIllegalValue(XKNXException):
    """Exception class for setting a value of a device with an illegal value."""

    def __init__(self, value, description):
        """Initialize DeviceIllegalValue class."""
        super(DeviceIllegalValue, self).__init__("Illegal value for device")
        self.value = value
        self.description = description

    def __str__(self):
        """Return object as readable string."""
        return '<DeviceIllegalValue description="{0}" value="{1}" />'.format(
            self.value,
            self.description)



"""Module for managing a KNX scene."""
from .device import Device
from .remote_value_scene_number import RemoteValueSceneNumber


class Scene(Device):
    """Class for managing a scene."""

    def __init__(self,
                 xknx,
                 name,
                 group_address=None,
                 scene_number=0,
                 device_updated_cb=None):
        """Initialize Sceneclass."""
        # pylint: disable=too-many-arguments
        super(Scene, self).__init__(xknx, name, device_updated_cb)

        self.scene_value = RemoteValueSceneNumber(
            xknx,
            group_address,
            device_name=self.name,
            after_update_cb=self.after_update)
        self.scene_number = int(scene_number)

    @classmethod
    def from_config(cls, xknx, name, config):
        """Initialize object from configuration structure."""
        group_address = \
            config.get('group_address')
        scene_number = \
            int(config.get('scene_number'))
        return cls(
            xknx,
            name=name,
            group_address=group_address,
            scene_number=scene_number)

    def has_group_address(self, group_address):
        """Test if device has given group address."""
        return self.scene_value.has_group_address(group_address)

    def __str__(self):
        """Return object as readable string."""
        return '<Scene name="{0}" ' \
               'scene_value="{1}" scene_number="{2}" />' \
            .format(
                self.name,
                self.scene_value.group_addr_str(),
                self.scene_number)

    async def run(self):
        """Move cover down."""
        await self.scene_value.set(self.scene_number)

    async def do(self, action):
        """Execute 'do' commands."""
        if action == "run":
            await self.run()
        else:
            self.xknx.logger.warning("Could not understand action %s for device %s", action, self.get_name())

    def state_addresses(self):
        """Return group addresses which should be requested to sync state."""
        return []

    def __eq__(self, other):
        """Equal operator."""
        return self.__dict__ == other.__dict__
